from typing import Callable, Tuple

from prometheus_client import Counter, Histogram, Summary
from starlette.requests import Request
from starlette.responses import Response


class Info:
    def __init__(
        self,
        request: Request,
        response: Response or None,
        method: str,
        modified_handler: str,
        modified_status: str,
        modified_duration: float,
    ):
        """Creates Info object that is used for instrumentation functions.

        This is the only argument that is passed to the instrumentation functions.

        Args:
            request: Python Requests request object.
            response Python Requests response object.
            method: Unmodified method of the request.
            modified_handler: Handler representation after processing by
                instrumentator. For example grouped to `none` if not templated.
            modified_status: Status code representation after processing by
                instrumentator. For example grouping into `2xx`, `3xx` and so on.
            modified_duration: Latency representation after processing by
                instrumentator. For example rounding of decimals. Seconds.
        """

        self.request = request
        self.response = response
        self.method = method
        self.modified_handler = modified_handler
        self.modified_status = modified_status
        self.modified_duration = modified_duration


def _build_label_attribute_names(
    should_include_handler: bool,
    should_include_method: bool,
    should_include_status: bool,
) -> Tuple[list, list]:
    """Builds up tuple with to be used label and attribute names.

    Args:
        should_include_handler: Should the `handler` label be part of the metric?
        should_include_method: Should the `method` label be part of the metric?
        should_include_status: Should the `status` label be part of the metric?

    Returns:
        Tuple with two list elements.

        First element: List with all labels to be used.
        Second element: List with all attribute names to be used from the
            `Info` object. Done like this to enable dynamic on / off of labels.
    """

    label_names = []
    info_attribute_names = []

    if should_include_handler:
        label_names.append("handler")
        info_attribute_names.append("modified_handler")

    if should_include_method:
        label_names.append("method")
        info_attribute_names.append("method")

    if should_include_status:
        label_names.append("status")
        info_attribute_names.append("modified_status")

    return label_names, info_attribute_names


# Metrics ======================================================================


def latency(
    metric_name: str = "http_request_duration_seconds",
    metric_doc: str = "Duration of HTTP requests in seconds",
    metric_namespace: str = "",
    metric_subsystem: str = "",
    should_include_handler: bool = True,
    should_include_method: bool = True,
    should_include_status: bool = True,
    buckets: tuple = Histogram.DEFAULT_BUCKETS,
) -> Callable[[Info], None]:
    """Default metric for the Prometheus FastAPI Instrumentator.

    Args:
        metric_name: Name of the metric to be created. Must be unique.
        metric_doc: Documentation of the metric.
        metric_namespace: Namespace of all  metrics in this metric function.
        metric_subsystem: Subsystem of all  metrics in this metric function.
        should_include_handler: Should the `handler` label be part of the metric?
        should_include_method: Should the `method` label be part of the metric?
        should_include_status: Should the `status` label be part of the metric?
        buckets: Buckets for the histogram. Defaults to Prometheus default.

    Returns:
        Function that takes a single parameter `Info`.
    """

    if buckets[-1] != float("inf"):
        buckets = buckets + (float("inf"),)

    label_names, info_attribute_names = _build_label_attribute_names(
        should_include_handler, should_include_method, should_include_status
    )

    if label_names:
        METRIC = Histogram(
            metric_name,
            metric_doc,
            labelnames=label_names,
            buckets=buckets,
            namespace=metric_namespace,
            subsystem=metric_subsystem,
        )
    else:
        METRIC = Histogram(
            metric_name,
            metric_doc,
            buckets=buckets,
            namespace=metric_namespace,
            subsystem=metric_subsystem,
        )

    def instrumentation(info: Info) -> None:
        if label_names:
            label_values = []
            for attribute_name in info_attribute_names:
                label_values.append(getattr(info, attribute_name))
            METRIC.labels(*label_values).observe(info.modified_duration)
        else:
            METRIC.observe(info.modified_duration)

    return instrumentation


def request_size(
    metric_name: str = "http_request_size_bytes",
    metric_doc: str = "Content bytes of requests.",
    metric_namespace: str = "",
    metric_subsystem: str = "",
    should_include_handler: bool = True,
    should_include_method: bool = True,
    should_include_status: bool = True,
) -> Callable[[Info], None]:
    """Record the content length of incoming requests.

    If content length is missing 0 will be assumed.

    Args:
        metric_name: Name of the metric to be created. Must be unique.
        metric_doc: Documentation of the metric.
        metric_namespace: Namespace of all  metrics in this metric function.
        metric_subsystem: Subsystem of all  metrics in this metric function.
        should_include_handler: Should the `handler` label be part of the metric?
        should_include_method: Should the `method` label be part of the metric?
        should_include_status: Should the `status` label be part of the metric?

    Returns:
        Function that takes a single parameter `Info`.
    """

    label_names, info_attribute_names = _build_label_attribute_names(
        should_include_handler, should_include_method, should_include_status
    )

    if label_names:
        METRIC = Summary(
            metric_name,
            metric_doc,
            labelnames=label_names,
            namespace=metric_namespace,
            subsystem=metric_subsystem,
        )
    else:
        METRIC = Summary(
            metric_name,
            metric_doc,
            namespace=metric_namespace,
            subsystem=metric_subsystem,
        )

    def instrumentation(info: Info) -> None:
        content_length = info.request.headers.get("Content-Length", 0)
        if label_names:
            label_values = []
            for attribute_name in info_attribute_names:
                label_values.append(getattr(info, attribute_name))
            METRIC.labels(*label_values).observe(int(content_length))
        else:
            METRIC.observe(int(content_length))

    return instrumentation


def response_size(
    metric_name: str = "http_response_size_bytes",
    metric_doc: str = "Content bytes of responses.",
    metric_namespace: str = "",
    metric_subsystem: str = "",
    should_include_handler: bool = True,
    should_include_method: bool = True,
    should_include_status: bool = True,
) -> Callable[[Info], None]:
    """Record the content length of outgoing responses.

    If content length is missing 0 will be assumed.

    Args:
        metric_name: Name of the metric to be created. Must be unique.
        metric_doc: Documentation of the metric.
        metric_namespace: Namespace of all  metrics in this metric function.
        metric_subsystem: Subsystem of all  metrics in this metric function.
        should_include_handler: Should the `handler` label be part of the metric?
        should_include_method: Should the `method` label be part of the metric?
        should_include_status: Should the `status` label be part of the metric?

    Returns:
        Function that takes a single parameter `Info`.
    """

    label_names, info_attribute_names = _build_label_attribute_names(
        should_include_handler, should_include_method, should_include_status
    )

    if label_names:
        METRIC = Summary(
            metric_name,
            metric_doc,
            labelnames=label_names,
            namespace=metric_namespace,
            subsystem=metric_subsystem,
        )
    else:
        METRIC = Summary(
            metric_name,
            metric_doc,
            namespace=metric_namespace,
            subsystem=metric_subsystem,
        )

    def instrumentation(info: Info) -> None:
        if info.response and hasattr(info.response, "headers"):
            content_length = info.response.headers.get("Content-Length", 0)
        else:
            content_length = 0

        if label_names:
            label_values = []
            for attribute_name in info_attribute_names:
                label_values.append(getattr(info, attribute_name))
            METRIC.labels(*label_values).observe(int(content_length))
        else:
            METRIC.observe(int(content_length))

    return instrumentation


def combined_size(
    metric_name: str = "http_combined_size_bytes",
    metric_doc: str = "Content bytes of requests and responses.",
    metric_namespace: str = "",
    metric_subsystem: str = "",
    should_include_handler: bool = True,
    should_include_method: bool = True,
    should_include_status: bool = True,
) -> Callable[[Info], None]:
    """Record the combined content length of requests and responses.

    If content length is missing 0 will be assumed.

    Args:
        metric_name: Name of the metric to be created. Must be unique.
        metric_doc: Documentation of the metric.
        metric_namespace: Namespace of all  metrics in this metric function.
        metric_subsystem: Subsystem of all  metrics in this metric function.
        should_include_handler: Should the `handler` label be part of the metric?
        should_include_method: Should the `method` label be part of the metric?
        should_include_status: Should the `status` label be part of the metric?

    Returns:
        Function that takes a single parameter `Info`.
    """

    label_names, info_attribute_names = _build_label_attribute_names(
        should_include_handler, should_include_method, should_include_status
    )

    if label_names:
        METRIC = Summary(
            metric_name,
            metric_doc,
            labelnames=label_names,
            namespace=metric_namespace,
            subsystem=metric_subsystem,
        )
    else:
        METRIC = Summary(
            metric_name,
            metric_doc,
            namespace=metric_namespace,
            subsystem=metric_subsystem,
        )

    def instrumentation(info: Info) -> None:
        request_cl = info.request.headers.get("Content-Length", 0)

        if info.response and hasattr(info.response, "headers"):
            response_cl = info.response.headers.get("Content-Length", 0)
        else:
            response_cl = 0

        content_length = int(request_cl) + int(response_cl)

        if label_names:
            label_values = []
            for attribute_name in info_attribute_names:
                label_values.append(getattr(info, attribute_name))
            METRIC.labels(*label_values).observe(int(content_length))
        else:
            METRIC.observe(int(content_length))

    return instrumentation


def default(
    metric_namespace: str = "",
    metric_subsystem: str = "",
    should_only_respect_2xx_for_highr: bool = False,
    latency_highr_buckets: tuple = (
        0.01,
        0.025,
        0.05,
        0.075,
        0.1,
        0.25,
        0.5,
        0.75,
        1,
        1.5,
        2,
        2.5,
        3,
        3.5,
        4,
        4.5,
        5,
        7.5,
        10,
        30,
        60,
    ),
    latency_lowr_buckets: tuple = (0.1, 0.5, 1),
) -> Callable[[Info], None]:
    """Contains multiple metrics to cover multiple things.

    Combines several metrics into a single function. Also more efficient than
    multiple separate instrumentation functions that do more or less the same.

    You get the following:

    * `http_requests_total` (`handler`, `status`, `method`): Total number of
        requests by handler, status and method.
    * `http_request_size_bytes` (`handler`): Total number of incoming
        content length bytes by handler.
    * `http_response_size_bytes` (`handler`): Total number of outgoing
        content length bytes by handler.
    * `http_request_duration_highr_seconds` (no labels): High number of buckets
        leading to more accurate calculation of percentiles.
    * `http_request_duration_seconds` (`handler`):
        Kepp the bucket count very low. Only put in SLIs.

    Args:
        metric_namespace: Namespace of all  metrics in this metric function.
        metric_subsystem: Subsystem of all  metrics in this metric function.
        should_only_respect_2xx_for_highr: Should the metric
            `http_request_duration_highr_seconds` only include latencies of
            requests / responses that have a status code starting with `2`?
        latency_highr_buckets: Buckets tuple for high res histogram. Can be
            large because no labels are used.
        latency_lowr_buckets: Buckets tuple for low res histogram. Should be
            very small as all possible labels are included.

    Returns:
        Function that takes a single parameter `Info`.
    """

    if latency_highr_buckets[-1] != float("inf"):
        latency_highr_buckets = latency_highr_buckets + (float("inf"),)

    if latency_lowr_buckets[-1] != float("inf"):
        latency_lowr_buckets = latency_lowr_buckets + (float("inf"),)

    TOTAL = Counter(
        name="http_requests_total",
        documentation="Total number of requests by method, status and handler.",
        labelnames=(
            "method",
            "status",
            "handler",
        ),
        namespace=metric_namespace,
        subsystem=metric_subsystem,
    )

    IN_SIZE = Summary(
        name="http_request_size_bytes",
        documentation=(
            "Content length of incoming requests by handler. "
            "Only value of header is respected. Otherwise ignored. "
            "No percentile calculated. "
        ),
        labelnames=("handler",),
        namespace=metric_namespace,
        subsystem=metric_subsystem,
    )

    OUT_SIZE = Summary(
        name="http_response_size_bytes",
        documentation=(
            "Content length of outgoing responses by handler. "
            "Only value of header is respected. Otherwise ignored. "
            "No percentile calculated. "
        ),
        labelnames=("handler",),
        namespace=metric_namespace,
        subsystem=metric_subsystem,
    )

    LATENCY_HIGHR = Histogram(
        name="http_request_duration_highr_seconds",
        documentation=(
            "Latency with many buckets but no API specific labels. "
            "Made for more accurate percentile calculations. "
        ),
        buckets=latency_highr_buckets,
        namespace=metric_namespace,
        subsystem=metric_subsystem,
    )

    LATENCY_LOWR = Histogram(
        name="http_request_duration_seconds",
        documentation=(
            "Latency with only few buckets by handler. "
            "Made to be only used if aggregation by handler is important. "
        ),
        buckets=latency_lowr_buckets,
        labelnames=("handler",),
        namespace=metric_namespace,
        subsystem=metric_subsystem,
    )

    def instrumentation(info: Info) -> None:
        TOTAL.labels(info.method, info.modified_status, info.modified_handler).inc()

        IN_SIZE.labels(info.modified_handler).observe(
            int(info.request.headers.get("Content-Length", 0))
        )

        if info.response and hasattr(info.response, "headers"):
            OUT_SIZE.labels(info.modified_handler).observe(
                int(info.response.headers.get("Content-Length", 0))
            )
        else:
            OUT_SIZE.labels(info.modified_handler).observe(0)

        if not should_only_respect_2xx_for_highr or info.modified_status.startswith("2"):
            LATENCY_HIGHR.observe(info.modified_duration)

        LATENCY_LOWR.labels(info.modified_handler).observe(info.modified_duration)

    return instrumentation



import gzip
import os
import re
from timeit import default_timer
from typing import Callable, Tuple

from fastapi import FastAPI
from prometheus_client import Gauge
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Match

from prometheus_fastapi_instrumentator import metrics


class PrometheusFastApiInstrumentator:
    def __init__(
        self,
        should_group_status_codes: bool = True,
        should_ignore_untemplated: bool = False,
        should_group_untemplated: bool = True,
        should_round_latency_decimals: bool = False,
        should_respect_env_var: bool = False,
        should_instrument_requests_inprogress: bool = False,
        excluded_handlers: list = [],
        round_latency_decimals: int = 4,
        env_var_name: str = "ENABLE_METRICS",
        inprogress_name: str = "http_requests_inprogress",
        inprogress_labels: bool = False,
    ):
        """Create a Prometheus FastAPI Instrumentator.

        Args:
            should_group_status_codes: Should status codes be grouped into
                `2xx`, `3xx` and so on?
            should_ignore_untemplated: Should requests without a matching
                template be ignored?
            should_group_untemplated: Should requests without a matching
                template be grouped to handler `none`?
            should_round_latency_decimals: Should recorded latencies be
                rounded to a certain number of decimals?
            should_respect_env_var: Should the instrumentator only work - for
                example the methods `instrument()` and `expose()` - if a
                certain environment variable is set to `true`? Usecase: A base
                FastAPI app that is used by multiple distinct apps. The apps
                only have to set the variable to be instrumented.
            should_instrument_requests_inprogress: Enables a gauge that shows
                the inprogress requests. See also the related args starting
                with `inprogress`.
            excluded_handlers: List of strings that will be compiled to regex
                patterns. All matches will be skipped and not instrumented.
            round_latency_decimals: Number of decimals latencies should be
                rounded to. Ignored unless `should_round_latency_decimals` is
                `True`.
            env_var_name: Any valid os environment variable name that will be
                checked for existence before instrumentation. Ignored unless
                `should_respect_env_var` is `True`.
            inprogress_name: Name of the gauge. Defaults to
                `http_requests_inprogress`. Ignored unless
                `should_instrument_requests_inprogress` is `True`.
            inprogress_labels: Should labels `method` and `handler` be part of
                the inprogress label? Ignored unless
                `should_instrument_requests_inprogress` is `True`.
        """

        self.should_group_status_codes = should_group_status_codes
        self.should_ignore_untemplated = should_ignore_untemplated
        self.should_group_untemplated = should_group_untemplated
        self.should_round_latency_decimals = should_round_latency_decimals
        self.should_respect_env_var = should_respect_env_var
        self.should_instrument_requests_inprogress = should_instrument_requests_inprogress

        self.round_latency_decimals = round_latency_decimals
        self.env_var_name = env_var_name
        self.inprogress_name = inprogress_name
        self.inprogress_labels = inprogress_labels

        if excluded_handlers:
            self.excluded_handlers = [re.compile(path) for path in excluded_handlers]
        else:
            self.excluded_handlers = []

        self.instrumentations = []

    def instrument(self, app: FastAPI) -> "self":
        """Performs the instrumentation by adding middleware.

        The middleware iterates through all `instrumentations` and execute them.

        Args:
            app: FastAPI app instance.

        Raises:
            e: Only raised if FastAPI itself throws an exception.

        Returns:
            self: Instrumentator. Builder Pattern.
        """

        if (
            self.should_respect_env_var
            and os.environ.get(self.env_var_name, "false") != "true"
        ):
            return self

        if len(self.instrumentations) == 0:
            self.instrumentations.append(metrics.default())

        self.inprogress = None
        if self.should_instrument_requests_inprogress:
            labels = (
                (
                    "method",
                    "handler",
                )
                if self.inprogress_labels
                else ()
            )
            self.inprogress = Gauge(
                name=self.inprogress_name,
                documentation="Number of HTTP requests in progress.",
                labelnames=labels,
                multiprocess_mode="livesum",
            )

        @app.middleware("http")
        async def dispatch_middleware(request: Request, call_next) -> Response:
            start_time = default_timer()

            handler, is_templated = self._get_handler(request)
            is_excluded = self._is_handler_excluded(handler, is_templated)
            handler = (
                "none" if not is_templated and self.should_group_untemplated else handler
            )

            if not is_excluded and self.should_instrument_requests_inprogress:
                if self.inprogress_labels:
                    inprogress = self.inprogress.labels(request.method, handler)
                else:
                    inprogress = self.inprogress
                inprogress.inc()

            try:
                response = None
                response = await call_next(request)
                status = str(response.status_code)
            except Exception as e:
                if response is None:
                    status = "500"
                raise e from None
            finally:
                if not is_excluded:
                    duration = max(default_timer() - start_time, 0)

                    if self.should_instrument_requests_inprogress:
                        inprogress.dec()

                    if self.should_round_latency_decimals:
                        duration = round(duration, self.round_latency_decimals)

                    if self.should_group_status_codes:
                        status = status[0] + "xx"

                    info = metrics.Info(
                        request=request,
                        response=response,
                        method=request.method,
                        modified_handler=handler,
                        modified_status=status,
                        modified_duration=duration,
                    )

                    for instrumentation in self.instrumentations:
                        instrumentation(info)

            return response

        return self

    def expose(
        self,
        app: FastAPI,
        should_gzip: bool = False,
        endpoint: str = "/metrics",
        include_in_schema: bool = True,
    ) -> "self":
        """Exposes endpoint for metrics.

        Args:
            app: FastAPI app instance. Endpoint will be added to this app.

            should_gzip: Should the endpoint return compressed data? It will
                also check for `gzip` in the `Accept-Encoding` header.
                Compression consumes more CPU cycles. In most cases it's best
                to just leave this option off since network bandwith is usually
                cheaper than CPU cycles. Defaults to `False`.

            endpoint: Endpoint on which metrics should be exposed.
            include_in_schema: Should the endpoint show up in the documentation?

        Raises:
            ValueError: If `prometheus_multiproc_dir` env var is found but
                doesn't point to a valid directory.

        Returns:
            self: Instrumentator. Builder Pattern.
        """

        if (
            self.should_respect_env_var
            and os.environ.get(self.env_var_name, "false") != "true"
        ):
            return self

        from prometheus_client import (
            CONTENT_TYPE_LATEST,
            REGISTRY,
            CollectorRegistry,
            generate_latest,
            multiprocess,
        )

        if "prometheus_multiproc_dir" in os.environ:
            pmd = os.environ["prometheus_multiproc_dir"]
            if os.path.isdir(pmd):
                registry = CollectorRegistry()
                multiprocess.MultiProcessCollector(registry)
            else:
                raise ValueError(
                    f"Env var prometheus_multiproc_dir='{pmd}' not a directory."
                )
        else:
            registry = REGISTRY

        @app.get(endpoint, include_in_schema=include_in_schema)
        def metrics(request: Request):
            if should_gzip and "gzip" in request.headers.get("Accept-Encoding", ""):
                return Response(
                    gzip.compress(generate_latest(registry)),
                    headers={"Content-Encoding": "gzip"},
                    media_type=CONTENT_TYPE_LATEST,
                )
            else:
                return Response(generate_latest(registry), media_type=CONTENT_TYPE_LATEST)

        return self

    def add(self, instrumentation_function: Callable[[metrics.Info], None]) -> "self":
        """Adds function to list of instrumentations.

        Args:
            instrumentation_function: Function that will be executed during
                every request handler call (if not excluded). See above for
                detailed information on the interface of the function.

        Returns:
            self: Instrumentator. Builder Pattern.
        """

        self.instrumentations.append(instrumentation_function)
        return self

    def _get_handler(self, request: Request) -> Tuple[str, bool]:
        """Extracts either template or (if no template) path.

        Args:
            request: Python Requests request object.

        Returns:
            Tuple with two elements.

            First element: Either template or if no template the path.
            Second element: If the path is templated or not.
        """

        for route in request.app.routes:
            match, child_scope = route.matches(request.scope)
            if match == Match.FULL:
                return route.path, True

        return request.url.path, False

    def _is_handler_excluded(self, handler: str, is_templated: bool) -> bool:
        """Determines if the handler should be ignored.

        Args:
            handler: Handler that handles the request.
            is_templated: Shows if the request is templated.

        Returns:
            `True` if excluded, `False` if not.
        """

        if is_templated is False and self.should_ignore_untemplated:
            return True

        if any(pattern.search(handler) for pattern in self.excluded_handlers):
            return True

        return False



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['prometheus_fastapi_instrumentator']

package_data = \
{'': ['*']}

install_requires = \
['fastapi==0.38.1', 'prometheus-client>=0.8.0,<0.9.0']

setup_kwargs = {
    'name': 'prometheus-fastapi-instrumentator',
    'version': '5.4.0',
    'description': 'Instrument your FastAPI with Prometheus metrics',
    'long_description': '# Prometheus FastAPI Instrumentator\n\n[![PyPI version](https://badge.fury.io/py/prometheus-fastapi-instrumentator.svg)](https://pypi.python.org/pypi/prometheus-fastapi-instrumentator/)\n[![Maintenance](https://img.shields.io/badge/maintained%3F-yes-green.svg)](https://GitHub.com/Naereen/StrapDown.js/graphs/commit-activity)\n[![downloads](https://pepy.tech/badge/prometheus-fastapi-instrumentator/month)](https://pepy.tech/project/prometheus-fastapi-instrumentator/month)\n[![docs](https://img.shields.io/badge/docs-here-blue)](https://trallnag.github.io/prometheus-fastapi-instrumentator/)\n\n![release](https://github.com/trallnag/prometheus-fastapi-instrumentator/workflows/release/badge.svg)\n![test branches](https://github.com/trallnag/prometheus-fastapi-instrumentator/workflows/test%20branches/badge.svg)\n[![codecov](https://codecov.io/gh/trallnag/prometheus-fastapi-instrumentator/branch/master/graph/badge.svg)](https://codecov.io/gh/trallnag/prometheus-fastapi-instrumentator)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nA configurable and modular Prometheus Instrumentator for your FastAPI. Install \n`prometheus-fastapi-instrumentator` from \n[PyPI](https://pypi.python.org/pypi/prometheus-fastapi-instrumentator/). Here \nis the fast track to get started with a preconfigured instrumentator:\n\n```python\nfrom prometheus_fastapi_instrumentator import Instrumentator\n\nInstrumentator().instrument(app).expose(app)\n```\n\nWith this, your FastAPI is instrumented and metrics ready to be scraped. The \nsensible defaults give you:\n\n* Counter `http_requests_total` with `handler`, `status` and `method`. Total \n    number of requests.\n* Summary `http_request_size_bytes` with `handler`. Added up total of the \n    content lengths of all incoming requests. If the request has no valid \n    content length, 0 bytes will be assumed.\n* Summary `http_response_size_bytes` with `handler`. Added up total of the \n    content lengths of all outgoing responses. If the response has no valid \n    content length, 0 bytes will be assumed.\n* Histogram `http_request_duration_seconds` with `handler`. Only a few buckets \n    to keep cardinality low. Use it for aggregations by handler or SLI buckets.\n* Histogram `http_request_duration_highr_seconds` without any labels. Large \n    number of buckets (>20) for accurate percentile calculations.\n\nIn addition, following behaviour is active:\n\n* Status codes are grouped into `2xx`, `3xx` and so on.\n* Requests without a matching template are grouped into the handler `none`.\n\n**If one of these presets does not suit your needs you can tweak \nbehaviour or register your own metric handler with the instrumentator. Find out\n[here](#creating-new-metrics) how to do that.**\n\n---\n\n> Don\'t hesitate to open an issue if you have questions or ideas regarding \n    the Prometheus FastAPI Instrumentator. Since I\'m currently using it in \n    production I\'m (still) motivated to work on it :wink:\n>  \n> -- <cite>trallnag</cite>\n\n---\n\nContents: **[Features](#features)** |\n**[Advanced Usage](#advanced-usage)** | \n[Creating the Instrumentator](#creating-the-instrumentator) |\n[Adding metrics](#adding-metrics) |\n[Creating new metrics](#creating-new-metrics) |\n[Perform instrumentation](#perform-instrumentation) |\n[Exposing endpoint](#exposing-endpoint) |\n**[Documentation](#documentation)** |\n**[Prerequesites](#prerequesites)** |\n**[Development](#development)**\n\n---\n\n## Features\n\nBeyond the fast track, this instrumentator is **highly configurable** and it \nis very easy to customize and adapt to your specific use case. Here is \na list of some of these options you may opt-in to:\n\n* Regex patterns to ignore certain routes.\n* Completely ignore untemplated routes.\n* Control instrumentation and exposition with an env var.\n* Rounding of latencies to a certain decimal number.\n* Renaming of labels and the metric.\n* Metrics endpoint can compress data with gzip.\n* Opt-in metric to monitor the number of requests in progress.\n\nIt also features a **modular approach to metrics** that should instrument all \nFastAPI endpoints. You can either choose from a set of already existing metrics \nor create your own. And every metric function by itself can be configured as \nwell. You can see ready to use metrics [here](https://trallnag.github.io/prometheus-fastapi-instrumentator/metrics.html).\n\n## Advanced Usage\n\nThis chapter contains an example on the advanced usage of the Prometheus \nFastAPI Instrumentator to showcase most of it\'s features. Fore more concrete \ninfo check out the \n[automatically generated documentation](https://trallnag.github.io/prometheus-fastapi-instrumentator/).\n\n### Creating the Instrumentator\n\nWe start by creating an instance of the Instrumentator. Notice the additional \n`metrics` import. This will come in handy later.\n\n```python\nfrom prometheus_fastapi_instrumentator import Instrumentator, metrics\n\ninstrumentator = Instrumentator(\n    should_group_status_codes=False,\n    should_ignore_untemplated=True,\n    should_respect_env_var=True,\n    should_instrument_requests_inprogress=True,\n    excluded_handlers=[".*admin.*", "/metrics"],\n    env_var_name="ENABLE_METRICS",\n    inprogress_name="inprogress",\n    inprogress_labels=True,\n)\n```\n\nUnlike in the fast track example, now the instrumentation and exposition will \nonly take place if the environment variable `ENABLE_METRICS` is `true` at \nrun-time. This can be helpful in larger deployments with multiple services\ndepending on the same base FastAPI.\n\n### Adding metrics\n\nLet\'s say we also want to instrument the size of requests and responses. For \nthis we use the `add()` method. This method does nothing more than taking a\nfunction and adding it to a list. Then during run-time every time FastAPI \nhandles a request all functions in this list will be called while giving them \na single argument that stores useful information like the request and \nresponse objects. If no `add()` at all is used, the default metric gets added \nin the background. This is what happens in the fast track example.\n\nAll instrumentation functions are stored as closures in the `metrics` module. \nClosures come in handy here because it allows us to configure the functions \nwithin.\n\n```python\ninstrumentator.add(metrics.latency(buckets=(1, 2, 3,)))\n```\n\nThis simply adds the metric you also get in the fast track example with a \nmodified buckets argument. But we would also like to record the size of \nall requests and responses. \n\n```python\ninstrumentator.add(\n    metrics.request_size(\n        should_include_handler=True,\n        should_include_method=False,\n        should_include_status=True,\n        metric_namespace="a",\n        metric_subsystem="b",\n    )\n).add(\n    metrics.response_size(\n        should_include_handler=True,\n        should_include_method=False,\n        should_include_status=True,\n        metric_namespace="namespace",\n        metric_subsystem="subsystem",\n    )\n)\n```\n\nYou can add as many metrics you like to the instrumentator.\n\n### Creating new metrics\n\nAs already mentioned, it is possible to create custom functions to pass on to\n`add()`. This is also how the default metrics are implemented. The \ndocumentation and code [here](https://trallnag.github.io/prometheus-fastapi-instrumentator/metrics.html) \nis helpful to get an overview.\n\nThe basic idea is that the instrumentator creates an `info` object that \ncontains everything necessary for instrumentation based on the configuration \nof the instrumentator. This includes the raw request and response objects \nbut also the modified handler, grouped status code and duration. Next, all \nregistered instrumentation functions are called. They get `info` as their \nsingle argument.\n\nLet\'s say we want to count the number of times a certain language \nhas been requested.\n\n```python\nfrom typing import Callable\nfrom prometheus_fastapi_instrumentator.metrics import Info\nfrom prometheus_client import Counter\n\ndef http_requested_languages_total() -> Callable[[Info], None]:\n    METRIC = Counter(\n        "http_requested_languages_total", \n        "Number of times a certain language has been requested.", \n        labelnames=("langs",)\n    )\n\n    def instrumentation(info: Info) -> None:\n        langs = set()\n        lang_str = info.request.headers["Accept-Language"]\n        for element in lang_str.split(",")\n            element = element.split(";")[0].strip().lower()\n            langs.add(element)\n        for language in langs:\n            METRIC.labels(language).inc()\n\n    return instrumentation\n```\n\nThe function `http_requested_languages_total` is used for persistent elements \nthat are stored between all instrumentation executions (for example the \nmetric instance itself). Next comes the closure. This function must adhere \nto the shown interface. It will always get an `Info` object that contains \nthe request, response and a few other modified informations. For example the \n(grouped) status code or the handler. Finally, the closure is returned.\n\n**Important:** The response object inside `info` can either be the response \nobject or `None`. In addition, errors thrown in the handler are not caught by \nthe instrumentator. I recommend to check the documentation and/or the source \ncode before creating your own metrics.\n\nTo use it, we hand over the closure to the instrumentator object.\n\n```python\ninstrumentator.add(http_requested_languages_total())\n```\n\n### Perform instrumentation\n\nUp to this point, the FastAPI has not been touched at all. Everything has been \nstored in the `instrumentator` only. To actually register the instrumentation \nwith FastAPI, the `instrument()` method has to be called.\n\n```python\ninstrumentator.instrument(app)\n```\n\nNotice that this will do nothing if `should_respect_env_var` has been set \nduring construction of the instrumentator object and the respective env var \nis not found.\n\n### Exposing endpoint\n\nTo expose an endpoint for the metrics either follow \n[Prometheus Python Client](https://github.com/prometheus/client_python) and \nadd the endpoint manually to the FastAPI or serve it on a separate server.\nYou can also use the included `expose` method. It will add an endpoint to the \ngiven FastAPI. With `should_gzip` you can instruct the endpoint to compress the \ndata as long as the client accepts gzip encoding. Prometheus for example does \nby default. Beware that network bandwith is often cheaper than CPU cycles.\n\n```python\ninstrumentator.expose(app, include_in_schema=False, should_gzip=True)\n```\n\nNotice that this will to nothing if `should_respect_env_var` has been set \nduring construction of the instrumentator object and the respective env var \nis not found.\n\n## Documentation\n\nThe documentation is hosted [here](https://trallnag.github.io/prometheus-fastapi-instrumentator/).\n\n## Prerequesites\n\n* `python = "^3.6"` (tested with 3.6 and 3.9)\n* `fastapi = ">=0.38.1, <=1.0.0"` (tested with 0.38.1 and 0.61.0)\n* `prometheus-client = "^0.8.0"` (tested with 0.8.0)\n\n## Development\n\nDeveloping and building this package on a local machine requires \n[Python Poetry](https://python-poetry.org/). I recommend to run Poetry in \ntandem with [Pyenv](https://github.com/pyenv/pyenv). Once the repository is \ncloned, run `poetry install` and `poetry shell`. From here you may start the \nIDE of your choice.\n\nTake a look at the Makefile or workflows on how to test this package.\n',
    'author': 'Tim Schwenke',
    'author_email': 'tim.and.trallnag+code@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/trallnag/prometheus-fastapi-instrumentator',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)
